"""runs sequences of commands."""
import subprocess
import json
from threading import Timer

from .console import console
from .configurer import configuration

SEPARATOR = "="

def find_sequence(idz, sequence):
    """find the sequence from sequence.json file."""
    sequence_id = idz[0]
    found = False

    if sequence_id in sequence:
        sequence = sequence[sequence_id]
        if len(idz) == 1:
            found = True
    else:
        console.print("Couldn't find the sequence", style='bad')
        return found, None
    for id_idx in range(1,len(idz)):
        sequence_id = idz[id_idx]

        for entry in sequence['seq']:
            found = False
            is_subseq = "id" in entry and sequence_id == entry["id"]
            is_command = "name" in entry and sequence_id == entry["name"]
            if is_subseq or is_command:
                if "id" in entry:
                    sequence = entry
                else:
                    one_element_seq = {'seq':[entry]}
                    sequence = one_element_seq
                found = True
                break
    return found, sequence

def run(args):
    """run the sequence from sequence.json file."""
    try:
        file = open (configuration['path'], "r", encoding="utf-8")
        default = json.load(file)
    except:
        console.print ("There is no sequence file I can find at the configured path.", style='bad')
        return

    sequence = default

    found, run_this_sequence = find_sequence(args.id, sequence)

    if not found:
        console.print ("I have nothing to run")
        return

    options = {'dryrun': False, 'failearly': args.failearly, 'verbose': args.verbose}

    with console.status("[blue]Running sequences...[/blue]\n"):
        try:
            run_sequence(run_this_sequence, args.params, options)
        except Exception as exception:
            console.print("An error occurred while running the sequences", style="bad")
            console.print(exception, style="bad")

def dryrun(args):
    """dryrun the sequence from sequence.json file."""
    try:
        file = open (configuration['path'], "r", encoding="utf-8")
        default = json.load(file)
    except Exception as exception:
        console.print ("There is no sequence file I can find at the configured path.", style='bad')
        console.print(exception, style="bad")
        return

    sequence = default
    found, run_this_sequence = find_sequence(args.id, sequence)

    if not found:
        console.print ("I have nothing to run", style="bad")
        return

    options = {'dryrun': True, 'failearly': True}

    try:
        run_sequence(run_this_sequence, args.params, options)
    except Exception as exception:
        console.print("An error occurred while running the sequences", style="bad")
        console.print(exception, style="bad")

def run_sequence(sub_seq, params, options):
    """start every process defined in the sequence."""
    if 'seq' not in sub_seq.keys() or sub_seq['seq'] is None:
        console.print ('Nothing to run', style='bad')
        return
    for entry in sub_seq['seq']:
        if "name" in entry.keys():
            console.print ('[accent]name[/accent]   :', entry["name"])
            comm = entry["command"]
            if params is not None and len(params) > 0:
                for parameter in params:
                    if SEPARATOR in parameter:
                        name, value = parameter.split(SEPARATOR, 1)
                        comm = comm.replace(name, value)

            console.print ('[accent]command[/accent]:', entry["command"])
            if not options['dryrun']:
                errcode = 0
                try:
                    with subprocess.Popen(comm, shell=True, stdout=subprocess.PIPE, bufsize=1, universal_newlines=True) as process:
                        timer = Timer(configuration['timeout'], stop_process,[process])
                        timer.start()
                        for line in process.stdout:
                            if options["verbose"] or configuration["_global_verbose"]:
                                console.print(line, end='', markup=False)
                finally:
                    timer.cancel()

                errcode = process.returncode
                if errcode != 0:
                    console.print("An error occurred while running the command!", style="bad")
                    if options["failearly"] or configuration["_global_failearly"]:
                        console.print('Fail early policy on, stopping execution.', style="bad")
                        return
                else:
                    console.print("Sequence completed!", style="good")

        elif "id" in entry.keys():
            run_sequence(entry, params, options)
        else:
            console.print ("Bad sequence definition??")

def stop_process(process):
    """stop the process if the timeout was exceeded."""
    console.print('Timeout expired, stopping process', style='bad')
    process.terminate()
    


"""configuration manager module."""
import ssl
import os
from configparser import ConfigParser
from pathlib import Path

from .utility import get_boolean
from .utility import get_int
from .utility import get_sequence_file_path
from .console import console

props = Path(os.path.dirname(__file__)) / "raumd.conf"
configurer = ConfigParser()
configurer.read(props)

configuration = {}
configuration["url"] = configurer['working']["url"]
configuration["path"] = get_sequence_file_path(configurer['working']["path"])
configuration["_global_failearly"] = get_boolean(configurer['working']["failearly"])
configuration["_global_verbose"] = get_boolean(configurer['working']["verbose"])
configuration["timeout"] = get_int(configurer['working']["timeout"])
configuration["localssl"] = get_boolean(configurer['working']["localssl"])

if configuration["localssl"]:
    configuration["context"] = ssl._create_unverified_context()
else:
    configuration["context"] = None

def configure(args):
    """global configuration"""
    if args.reset:

        url = 'https://airlocks.xyz'
        path = 'sequence.json'
        timeout = ''
        failearly = 'Yes'
        localssl = 'Yes'
        verbose = 'Yes'

        if not configurer.has_section('working'):
            configurer.add_section('working')

        configurer.set("working", "url", url)
        configurer.set("working", "path", path)
        configurer.set("working", "timeout", timeout)
        configurer.set("working", "failearly", failearly)
        configurer.set("working", "localssl", localssl)
        configurer.set("working", "verbose", verbose)

    elif args.show:

        console.print("[accent]url[/accent]      :", configurer['working']["url"])
        console.print("[accent]path[/accent]     :", configurer['working']["path"])
        console.print("[accent]timeout[/accent]  :", configurer['working']["timeout"])
        console.print("[accent]localssl[/accent] :", configurer['working']['localssl'])
        console.print("[accent]verbose[/accent]  :", configurer['working']["verbose"])
        console.print("[accent]failearly[/accent]:", configurer['working']["failearly"])

    else:
        if args.url is not None:
            console.print('Setting url to: ', args.url[0])
            configurer.set("working", "url", args.url[0])

        if args.path is not None:
            console.print('Setting path to: ', args.path[0])
            configurer.set("working", "path", args.path[0])

        if args.timeout is not None:
            console.print('Setting timeout to: ', args.timeout[0])
            configurer.set("working", "timeout", str(args.timeout[0]))

        if args.failearly is not None:
            console.print('Setting failearly to: ', args.failearly[0])
            configurer.set("working", "failearly", args.failearly[0])

        if args.localssl is not None:
            console.print('Setting localssl to: ', args.localssl[0])
            configurer.set("working", "localssl", args.localssl[0])

        if args.verbose is not None:
            console.print('Setting verbose to: ', args.verbose[0])
            configurer.set("working", "verbose", args.verbose[0])

    if not args.show:
        try:
            with open(props, 'w', encoding="utf-8") as configfile:
                configurer.write(configfile)
        except Exception as exception:
            console.print('An exception occurred while trying to write to the configure file.',
                style='bad')
            console.print(exception, style='bad')



"""captures the exit signal."""
import signal
from .console import console

class GracefulExiter():
    """Helper class that intercepts the interrupt signal."""

    def __init__(self):
        """Object init."""
        signal.signal(signal.SIGINT, self.change_state)

    def change_state(self, signum, frame):
        """Print message, stop processing."""
        console.print("Received interrupt signal. Exitting now!", style='bad')
        signal.signal(signal.SIGINT, signal.SIG_DFL)



"""utility function library."""
import os
from pathlib import Path
import json
from .console import console

GET_INT_EXCEPTION = 'An exception occurred while getting the value for an integer parameter'

def get_boolean(param):
    """boolean value from string"""
    ret = param in ('Yes', 'Y', 'True', '1', True, 1)
    return ret

def get_int(param):
    """obtain integer value from string"""
    value = None
    if (not param or param == '0'):
        value = None
    else:
        try:
            value = int(param)
        except Exception as exception:
            console.print(GET_INT_EXCEPTION)
            console.print(exception)
    return value

def get_sequence_file_path(ipath):
    """obtain path for sequence file, helpful for relative file paths"""
    opath = None
    if not os.path.isabs(ipath):
        opath = Path(os.getcwd()) / ipath
    else:
        opath = ipath

    return opath

def write_sequence_file(path):
    """write sequence file"""
    if not os.path.exists(path):
        with open(path, "w", encoding="utf-8") as file:
            empty = {}
            json.dump(empty,file)



"""imports commands into the sequence file."""
import json
import urllib.request as urllib2

from .configurer import configuration
from .console import console
from .utility import write_sequence_file

def download(args):
    """downloads sequence."""
    write_sequence_file(configuration['path'])

    with open (configuration['path'], "r", encoding="utf-8") as file:
        default = json.load(file)

        status, sequence = download_sequence(args.id)

        if status == "ok":
            with open (configuration['path'], "w", encoding="utf-8") as sequence_file:
                new_sequence_name = args.id[len(args.id) -1]
                if args.rename is not None:
                    new_sequence_name = args.rename[0]
                default[new_sequence_name] = sequence
                json.dump(default, sequence_file)
                console.print("Import succeded.", style="good")
        else:
            console.print('Couldn\'t find the sequence:', style='bad')
            console.print(sequence)

def download_sequence(sequence):
    """downloads sequence."""
    try:
        body = {"sub_seq": []}
        if len(sequence) > 1:
            for entry in sequence[1:len(sequence)]:
                body.get('sub_seq').append(entry)
        jsondata = json.dumps(body)
        jsondataasbytes = jsondata.encode('utf-8')
        req = urllib2.Request(configuration['url'] + "/api/download/" + sequence[0])

        req.add_header('Content-Length', len(jsondataasbytes))
        req.add_header('Content-Type', 'application/json')

        with urllib2.urlopen(req, jsondataasbytes, context=configuration['context']) as response:
            text = response.read()
            stud_obj = json.loads(text)
            return "ok", stud_obj

    except Exception as exception:
        console.print("Oops download failed!", style="bad")
        console.print(exception, style="bad")
        return "nok", None



"""script entrypoint. defines arguments, assigns operations to modules."""

import argparse
import sys

from .__init__ import __version__
from .runner import run
from .runner import dryrun
from .importer import download
from .configurer import configure
from .inventory import show
from .graceful_exiter import GracefulExiter

RUN_HELP = 'Runs a command or a sequence.'
RUN_ID_HELP = 'The sequence or command to be executed.'
RUN_PARAMS = 'The parameters to run the sequences with.'
RUN_FAIL_EARLY = 'If present stops on first command failure.'
RUN_VERBOSE = 'Print the output of commands.'

DOWNLOAD_HELP = 'Downloads a sequence from airlocks.xyz'
DOWNLOAD_ID_HELP = 'The sequence to obtain.'
DOWNLOAD_RENAME_HELP = 'Renames the sequence for local use.'

DRY_RUN_HELP = 'Dryruns a command or a sequence.'

SHOW_HELP = 'Shows the existing commands and/or a sequence.'
SHOW_ID_HELP = 'The sequence or command to be executed.'

CONFIG_HELP = 'Configures the runner.'
CONFIG_URL_HELP = 'Sets the download url.'
CONFIG_PATH_HELP = 'Path to json file holding the sequences.'
CONFIG_LOCALSSL_HELP = 'Accept locally generated certificates. (Yes/No)'
CONFIG_TIMEOUT_HELP = 'Timeout for every command in seconds.'
CONFIG_FAILEARLY_HELP = 'General exit strategy. (Yes/No)'
CONFIG_RESET_HELP = 'Reset configuration to defaults.'
CONFIG_SHOW_HELP = 'Show current configuration.'
CONFIG_VERBOSE_HELP = 'Print the output of commands. (Yes/No)'

def main():
    """main function"""
    description = """Raumdeuter runs your sequence of bash commands defined in a json file."""
    parser = argparse.ArgumentParser(description = description)
    subparsers = parser.add_subparsers()

    run_parser = subparsers.add_parser('run',
        help = RUN_HELP)
    run_parser.add_argument("id", type = str, nargs = '+',
        help = RUN_ID_HELP)
    run_parser.add_argument("--params", type = str, nargs = '*',
        help = RUN_PARAMS)
    run_parser.add_argument("--failearly", "-failearly", "-fe", "--fe", action='store_true',
        help = RUN_FAIL_EARLY)
    run_parser.add_argument('--verbose', '-verbose', '-v', action='store_true',
        help = RUN_VERBOSE)
    run_parser.set_defaults(func=run)

    download_parser = subparsers.add_parser('download',
        help = DOWNLOAD_HELP)
    download_parser.add_argument("id", type = str, nargs = '+',
        help = DOWNLOAD_ID_HELP)
    download_parser.add_argument("--rename", type = str, nargs = '*',
        help = DOWNLOAD_RENAME_HELP)
    download_parser.set_defaults(func=download)

    dryrun_parser = subparsers.add_parser('dryrun',
        help = DRY_RUN_HELP)
    dryrun_parser.add_argument("id", type = str, nargs = '+',
        help = RUN_ID_HELP)
    dryrun_parser.add_argument("--params", type = str, nargs = '*',
        help = RUN_PARAMS)
    dryrun_parser.set_defaults(func=dryrun)

    show_parser = subparsers.add_parser('show',
        help = SHOW_HELP)
    show_parser.add_argument("id", type = str, nargs = '*',
        help = SHOW_ID_HELP)
    show_parser.set_defaults(func=show)

    configure_parser = subparsers.add_parser('configure',
        help = CONFIG_HELP)
    configure_parser.add_argument("--url", "-u", "-url", "--download-url", "-download-url",
        type = str, nargs = 1,
        help = CONFIG_URL_HELP)
    configure_parser.add_argument("--path", "-p", "-path", "--download-path", "-download-path",
        type = str, nargs = 1,
        help = CONFIG_PATH_HELP)
    configure_parser.add_argument("--localssl", "-localssl",
        type = str, nargs = 1,
        help = CONFIG_LOCALSSL_HELP)
    configure_parser.add_argument("--timeout", "-t",
        type = str, nargs = 1,
        help = CONFIG_TIMEOUT_HELP)
    configure_parser.add_argument("--failearly", "-fe",
        type = str, nargs = 1,
        help = CONFIG_FAILEARLY_HELP)
    configure_parser.add_argument("--reset", "-r",
        action='store_true',
        help = CONFIG_RESET_HELP)
    configure_parser.add_argument("--show", "-show",
        action='store_true',
        help = CONFIG_SHOW_HELP)
    configure_parser.add_argument('--verbose', '-verbose',
        type = str, nargs = 1,
        help = CONFIG_VERBOSE_HELP)
    configure_parser.set_defaults(func=configure)

    parser.add_argument("-v", "--v", "-version", "--version", action = 'version',
        version=__version__, default = None, help = "console prints the version of the script.")

    if len(sys.argv) <= 1:
        sys.argv.append('--help')

    args = parser.parse_known_args()

    args[0].func(args[0])

flag = GracefulExiter()

if __name__ == "__main__":
    main()



"""lists the sequences and available commands."""
import json
from .configurer import configuration
from .console import console
from .utility import write_sequence_file

def show(args):
    """finds and prints the requested sequence"""
    write_sequence_file(configuration['path'])

    try:
        with open (configuration['path'], "r", encoding="utf-8") as file:
            default = json.load(file)
    except Exception as exception:
        console.print(exception, style='bad')
        console.print ("There is no sequence file I can find at the configured path.", style='bad')
        return

    sequence = default
    idz = args.id

    if len(idz) == 0:
        one_element_seq = {'seq':[]}
        for entry in sequence:
            one_element_seq.get('seq').append({'id':entry})
        show_sequence(one_element_seq)
        return
        
    sequence_id = idz[0]

    found = False
    last_found_id = None

    if sequence_id in sequence:
        last_found_id = sequence_id
        sequence = sequence[sequence_id]
        found = True
    for id_idx in range(1,len(idz)):
        sequence_id = idz[id_idx]

        for entry in sequence['seq']:
            found = False
            is_subseq = "id" in entry and sequence_id == entry["id"]
            is_command = "name" in entry and sequence_id == entry["name"]
            if is_subseq or is_command:
                last_found_id = sequence_id
                if "id" in entry:
                    sequence = entry
                else:
                    one_element_seq = {'seq':[entry]}
                    sequence = one_element_seq
                found = True
                break

    if not found:
        console.print ("Last found id: ", last_found_id)
        return
    show_sequence(sequence)

def show_sequence(sequence):
    """prints the sequence"""
    seq = sequence.get('seq')
    if seq is None or len(seq) == 0:
        console.print("There are no sequences.", style='good')
    try:
        for entry in seq:
            if "name" in entry:
                console.print("[accent]name[/accent]   : "
                    + "[good]" + entry["name"] + "[/good]")
                console.print("[accent]command[/accent]: " + "[good]"
                    + entry["command"] + "[/good]")
            elif "id" in entry:
                console.print("[accent]subseq[/accent]: "
                    + "[important]" + entry["id"] + "[/important]" +
                    (" / [important]" + entry["title"]
                    + "[/important]" if "title" in entry else ""))
    except Exception as exception:
        console.log("An error occurred while looking at the entry", style='bad')
        console.print(exception, style="bad")



"""console customizer."""
from rich.console import Console
from rich.theme import Theme

custom_theme = Theme({
    "good" : "green",
    "bad": "red",
    "important": "purple",
    "accent": "blue",
    })
console = Console(theme=custom_theme)



"""raumd installer."""

from pathlib import Path
from setuptools import setup, find_packages
from main.__init__ import __version__

with open(Path('main') / 'requirements.txt', encoding="utf-8") as f:
    requirements = f.readlines()

LONG_DESCRIPTION = 'Raumdeuter runs your sequence of bash commands defined custom json files. \
    This files can be created on the online platform https://airlocks.xyz.'

setup(
        name ='raumd',
        version =__version__,
        author ='Roman Botnari',
        author_email ='romanbotnari@me.com',
        url ='',
        description ='.',
        long_description = LONG_DESCRIPTION,
        long_description_content_type ="text/markdown",
        license ='Apache License, Version 2.0',
        packages = find_packages(),
        entry_points = {
            'console_scripts': [
                'raumd = main.raumd:main'
            ]
        },
        classifiers =[
            "Programming Language :: Python :: 3",
            "Operating System :: OS Independent",
        ],
        keywords ='',
        install_requires = requirements,
        zip_safe = False,
        package_data={
            "main": ["raumd.conf", "requirements.txt"],
        }
)
